

import argparse
from net.skip_model_wavelet_L0_4 import Rnet_wavelet, Diff
from net.skip_model_wavelet_L0_4 import L0 as L0_v4
from net.skip_model_wavelet_L0_5 import L0 as L0_v5
from net.skip_model import skip
from net.losses import  WeightedGradientLoss, Gradient_L0Loss,Gradient_L0LossV2,StructureLoss,Gradient_L0LossV3
from net.noise import get_noise
from utils.image_io import *
from utils.LambdaLR import *
# from skimage.measure import compare_psnr
from skimage.metrics import peak_signal_noise_ratio as compare_psnr
import numpy as np
import torch
import torch.nn as nn
from collections import namedtuple
from torchvision.utils import save_image as save_tensor

import time
import sys
import os
import matplotlib.pyplot as plt
import math

class Separation(object):
    def __init__(self, opt, image_name, image, plot_during_training=True, show_every=5000, num_iter=20000,
                 original_reflection=None, original_transmission=None):
        self.opt = opt
        self.image = image
        self.plot_during_training = plot_during_training
        self.psnrs = []
        self.show_every = show_every
        self.image_name = image_name
        self.num_iter = num_iter
        self.loss_function = None
        self.parameters = None
        self.reflectance_net_inputs = None
        self.shading_net_inputs = None
        self.original_transmission = original_transmission
        self.original_reflection = original_reflection
        self.reflectance_net = None
        self.shading_net = None
        self.total_loss = None
        self.reflectance_out = None
        self.shading_out = None
        self.current_result = None
        self.best_result = None
        self.total_losses = []
        self.recon_losses = []
        self.grad_d_losses = []
        self.grad_s_losses = []
        self.Rsmooth_losses = []
        self.R_grad_l0_losses = []
        self.R_grad_mask_ssim_losses = []
        self.S_grad_l0_losses = []
        self.mask_weight_losses = []
        self.width = None
        self.height = None
        self.originalroot = 'output-p1/'
        self.outputroot = 'output-p1/'
        self.start_decay = 4000
        self.eps = 1e-5
        self.w_Rsmooth = 0.2 
        self.w_grad = 0.1
        self.w_grad_smooth = 0.1
        self.threshold = 0.02
        self.alpha = 1.2
        self.center = False
        self.init_weight = True
        self.R_weight = None
        self.learning_rate = 0.002
        self.input_depth = 2
        self.in_shading_depth = 2              
        self.kernel_size = 5
        self.R_skip = [0,0,1,1,0]
        self.S_skip = [0,0,1,1,0]
        self.R_pool_H = [2,2,2,2,2]
        self.R_pool_L = [2,2,2,2,2]
        self.S_pool_H = [2,2,2,2,2]
        self.S_pool_L = [2,2,2,2,2]
        self.n_seg = None
        self.compactness = None
        self.w_mask_l0 = 0.001
        self.w_grad_l0 = 1
        self.w_mask_weight = 0.1
        self._lambda = 2e-2
        self.smooth_start = 1000
        self.loc_mean = 1
        self.beta = 2./3.
        self.gamma = -0.3
        self.zeta = 1.3
        self.ifdetach = False
        self.temp_L0mask = 0
        self.dif_L0mask = 0
        self.count = 0
        self.dif_L0mask_record = 1
        self.start_pre = 0
        self.w_Sgrad_l0 = 1
        self.w_Rgrad_l0_ssim = 1
        self.ssim_win_size = 11
        self.metric = 'SSIM'
        self.mask_threshold = 0.3
        self.w_mask_R_weight = 3
        self.w_mask_R_l0 = 1
        self.mask_iters = 500
        self.net_iters = 50
        self.R_mask_start = 2000
        self.w_Rgrad_l0_comple = 0.02

    def _init_all(self):
        self._init_images()
        self._init_nets()
        self._init_inputs()
        self._init_parameters()
        self._init_losses()

    def _init_images(self):
        self.images = create_augmentations(self.image)
        self.images_torch = [np_to_torch(image).type(torch.cuda.FloatTensor) for image in self.images]
        if self.opt.param_mode == 'MIT':
            self.weight_for_temp_penalty, self.weight_for_mask = self.compute_Weight2(self.images_torch[0], self.alpha)
        else:
            self.weight_for_mask = self.compute_Weight(self.images_torch[0], self.alpha)
        
        

    def _init_inputs(self): 
        input_type = 'noise'
        # input_type = 'meshgrid'
        data_type = torch.cuda.FloatTensor
        origin_noise = torch_to_np(get_noise(self.input_depth,
                                                  input_type,
                                                  (self.images_torch[0].shape[2],
                                                   self.images_torch[0].shape[3])).type(data_type).detach())
        self.reflectance_net_inputs = [np_to_torch(aug).type(data_type).detach() for aug in create_augmentations(origin_noise)]
        origin_noise = torch_to_np(get_noise(self.in_shading_depth,
                                             input_type,
                                             (self.images_torch[0].shape[2],
                                              self.images_torch[0].shape[3])).type(data_type).detach())
        self.shading_net_inputs = [np_to_torch(aug).type(data_type).detach() for aug in
                                      create_augmentations(origin_noise)]

    def _init_parameters(self):
        self.parameters = [p for p in self.reflectance_net.parameters()] + \
                          [p for p in self.shading_net.parameters()] + \
                          [p for p in self.L0_process.parameters()]

    def _init_nets(self):
        data_type = torch.cuda.FloatTensor
        pad = 'reflection'
        #Rnet_wavelet
        reflectance_net = Rnet_wavelet(
            self.images_torch[0].size(),
            self.input_depth, 3,
            num_channels_down=[8, 16, 32, 64, 128],
            num_channels_up=[8, 16, 32, 64, 128],
            num_channels_skip= self.R_skip,
            pool_H=self.R_pool_H, pool_L=self.R_pool_L,
            # n_seg =self.n_seg, compactness = self.compactness,
            
            upsample_mode='bilinear',
            filter_size_down=self.kernel_size,
            filter_size_up=self.kernel_size,
            need_sigmoid=True, need_bias=True, pad=pad, act_fun='LeakyReLU',
            loc_mean = self.loc_mean, loc_sdev = 0.01, beta = self.beta, gamma = self.gamma, zeta=self.zeta,fix_temp=True,)

        self.reflectance_net = reflectance_net.type(data_type)

##################################################################################################        
        
        shading_net = skip(
            self.in_shading_depth, 1,
            num_channels_down=[8, 16, 32, 64, 128],
            num_channels_up=[8, 16, 32, 64, 128],
            num_channels_skip=self.S_skip,

            upsample_mode='bilinear',
            filter_size_down=self.kernel_size,
            filter_size_up=self.kernel_size,
            need_sigmoid=True, need_bias=True, pad=pad, act_fun='LeakyReLU')

        self.shading_net = shading_net.type(data_type)

        if self.opt.param_mode == 'MIT':
            self.L0_process = L0_v5(self.images_torch[0].size(),
            loc_mean = self.loc_mean, loc_sdev = 0.01, beta = self.beta, gamma = self.gamma, zeta=self.zeta,fix_temp=True).type(data_type)
        else:
            self.L0_process = L0_v4(self.images_torch[0].size(),
            loc_mean = self.loc_mean, loc_sdev = 0.01, beta = self.beta, gamma = self.gamma, zeta=self.zeta,fix_temp=True).type(data_type)

    def _init_losses(self):
        data_type = torch.cuda.FloatTensor
        self.l1_loss = nn.L1Loss().type(data_type)
        self.mse_loss = torch.nn.MSELoss().type(data_type)
        self.grad_loss = WeightedGradientLoss(self.threshold, self.alpha).type(data_type)
        self.R_grad = Gradient_L0Loss().type(data_type)
        self.S_grad = Gradient_L0LossV2().type(data_type)
        self.R_grad_comple = Gradient_L0LossV3().type(data_type)
        self.R_grad_mask_ssim = StructureLoss(self.ssim_win_size,self.metric).type(data_type)


    def optimize(self):
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True
        optimizer = torch.optim.Adam(self.parameters, lr=self.learning_rate)

        while(self.dif_L0mask_record > self.start_pre or self.count<10):
            optimizer.zero_grad()
            self._optimization_pre()
            optimizer.step()
        
        self.count = 0
        for j in range(self.num_iter):
            
            if(j>=self.R_mask_start and j%self.net_iters==0):
                self.if_compute = True
                if self.opt.param_mode != 'MIT': self.weight_for_mask = self.compute_Weight(self.reflectance_out.detach(), self.alpha)
                sys.stdout.flush()
                for _ in range(self.mask_iters):
                    optimizer.zero_grad()
                    self._optimization_Rmask()
                    optimizer.step()

            optimizer.zero_grad()
            self._optimization_closure(j)                 
            self._obtain_current_result(j)
            
            print("Training Iteration %d/%d PSNR=%.2f"%(j, self.num_iter, self.current_result.psnr), '\r', end='')
            if self.plot_during_training and (j+1) % self.show_every == 0:
                self._plot_closure(j)
            
            optimizer.step()

    def _plot_closure(self, j):

        save_graph('{}_PSNR'.format(self.image_name), self.psnrs, self.outputroot)
        save_image('{}_R_Iter{}'.format(self.image_name, j), self.best_result.reflection, self.width,self.height,self.outputroot)
        save_image('{}_S_Iter{}'.format(self.image_name, j), self.best_result.transmission, self.width,self.height,self.outputroot)
        save_image('{}_Original'.format(self.image_name), self.images[0], self.width,self.height,self.outputroot)
        

    def _get_augmentation(self, iteration):
  
        return 0 

    def _optimization_pre(self):
        if self.opt.param_mode == 'MIT':
            self.l0_mask, self.l0_penalty, self.mask_weight = self.L0_process(self.images_torch[0], self.weight_for_temp_penalty, self.ifdetach)
        else:
            self.l0_mask, self.l0_penalty, self.mask_weight = self.L0_process(self.images_torch[0], self.alpha, self.images_torch[0], self.ifdetach)
        self.mask_weight_loss = self.l1_loss(self.weight_for_mask, self.l0_mask * self.weight_for_mask) * self.w_mask_weight
        self.l0 = self.l0_penalty * self.w_mask_l0
        self.total_loss =  self.l0 + self.mask_weight_loss
        self.total_loss.backward()
        
        if (self.count == 0):
            self.temp_L0mask = self.l0_mask
        self.dif_L0mask = torch.sign(((self.temp_L0mask - self.l0_mask).abs() - 0.3).clamp(min=0)).sum()
        self.dif_L0mask = self.dif_L0mask / torch.numel(self.l0_mask)
        self.temp_L0mask = self.l0_mask
        if self.dif_L0mask < self.dif_L0mask_record or self.count<10:
            self.dif_L0mask_record = self.dif_L0mask
        self.count = self.count + 1
        self.Rsmooth_losses.append(self.l0)
        self.mask_weight_losses.append(self.mask_weight_loss)
        if self.count % 100==0:
            print('PreIteration {:5d}   MaskChangeRate {:3f} '.format(self.count, self.dif_L0mask.item() ),'\r', end='')

    def _optimization_Rmask(self):
        if self.opt.param_mode == 'MIT':
            self.l0_mask, self.l0_penalty, self.mask_weight = self.L0_process(self.images_torch[0], self.weight_for_temp_penalty, self.ifdetach)
        else:
            self.l0_mask, self.l0_penalty, self.mask_weight = self.L0_process(self.reflectance_out.detach(), self.alpha, self.reflectance_out.detach(), self.ifdetach, self.if_compute)
        self.mask_weight_loss = self.l1_loss(self.weight_for_mask, self.l0_mask * self.weight_for_mask) * self.w_mask_R_weight
        self.l0 = self.l0_penalty * self.w_mask_R_l0
        self.total_loss =  self.l0 + self.mask_weight_loss
        self.total_loss.backward()

        if (self.count == 0):
            self.temp_L0mask = self.l0_mask
        self.dif_L0mask = torch.sign(((self.temp_L0mask - self.l0_mask).abs() - 0.3).clamp(min=0)).sum()
        self.dif_L0mask = self.dif_L0mask / torch.numel(self.l0_mask)
        self.temp_L0mask = self.l0_mask
        if self.dif_L0mask < self.dif_L0mask_record or self.count<10:
            self.dif_L0mask_record = self.dif_L0mask
        self.count = self.count + 1
        self.Rsmooth_losses.append(self.l0)
        self.mask_weight_losses.append(self.mask_weight_loss)
        if self.count % 100==0:
            print('R_maskIteration {:5d}   MaskChangeRate {:3f} '.format(self.count, self.dif_L0mask.item() ),'\r', end='')
        self.if_compute = False                     

    def _optimization_closure(self, step):
        if step == self.num_iter - 1:
            reg_noise_std = 0
        elif step < 1000:
            reg_noise_std = (1 / 1000.) * ((step) // 100)
            # reg_noise_std = (1 / 1000.)
        else:
            reg_noise_std = 1 / 1000.
        aug = self._get_augmentation(step)                     
        if step == self.num_iter - 1:
            aug = 0

        #reg_noise_std = 0    
        reflectance_net_input = self.reflectance_net_inputs[aug] + (self.reflectance_net_inputs[aug].clone().normal_() * reg_noise_std)
        shading_net_input = self.shading_net_inputs[aug] + (self.shading_net_inputs[aug].clone().normal_() * reg_noise_std)
        self.reflectance_out = self.reflectance_net(reflectance_net_input)
        self.shading_out = self.shading_net(shading_net_input)

        self.recon = self.l1_loss( torch.log(self.reflectance_out.clamp(min=self.eps)) + torch.log(self.shading_out.clamp(min=self.eps)),\
                                    torch.log(self.images_torch[aug].clamp(min=self.eps)))
        
        if self.smooth_start is None:
            lam = self.w_grad_l0
            lam2 = self.w_Rgrad_l0_ssim
            lam3 = self.w_Sgrad_l0
            
        else:
            if step < self.smooth_start:
                lam = 0
                lam2 = 0
                lam3 = 0

            else:
                self.increase = self.w_grad_l0 / 10
                lam = min(self.w_grad_l0, self.increase * math.ceil((step-self.smooth_start)/100))
                
                self.increase2 = self.w_Rgrad_l0_ssim / 10
                lam2 = min(self.w_Rgrad_l0_ssim, self.increase2 * math.ceil((step-self.smooth_start)/100))
                
                self.increase3 = self.w_Sgrad_l0 /10
                lam3 = min(self.w_Sgrad_l0, self.increase3 * math.ceil((step-self.smooth_start)/100))
                
        

        self.mask = self.l0_mask.detach()
        self.R_grad_comple_loss = self.R_grad_comple(self.reflectance_out, self.l0_mask.detach())
        self.R_grad_l0 = self.R_grad(self.reflectance_out,  self.mask.detach())   
        self.R_grad_mask_ssim_loss = self.R_grad_mask_ssim(self.reflectance_out, self.mask.detach())
        self.S_grad_l0 = self.S_grad(self.shading_out, self.mask.detach()) 

        self.grad_d, self.grad_smooth =  self.grad_loss(self.images_torch[0],self.reflectance_out, self.shading_out)
        self.grad_smooth = self.grad_smooth 
        self.grad_d = self.grad_d 
         

        self.total_loss = self.recon + (self.grad_d * self.w_grad + self.grad_smooth * self.w_grad_smooth) \
                         + self.R_grad_l0*lam + self.S_grad_l0*lam3 \
                         + self.R_grad_mask_ssim_loss * lam2 \
                         + self.R_grad_comple_loss * self.w_Rgrad_l0_comple
        self.total_loss.backward()

        self.total_losses.append(self.total_loss)
        self.recon_losses.append(self.recon)
        self.R_grad_l0_losses.append(self.R_grad_l0)
        self.R_grad_mask_ssim_losses.append( -self.R_grad_mask_ssim_loss )
        self.S_grad_l0_losses.append(self.S_grad_l0)
        self.grad_d_losses.append(self.grad_d)
        self.grad_s_losses.append(self.grad_smooth)

    def compute_Weight(self, image, alpha):
        dif = Diff()
        weight_h = dif(image,'h').abs().clamp(min=0.01)
        weight_w = dif(image,'w').abs().clamp(min=0.01)
        weight = (weight_h + weight_w)/2
        maximum = weight.max()
        minimum = weight.min()
        weight = (weight - minimum)/ (maximum - minimum)
        return weight

    def compute_Weight2(self, image, alpha):
        weight_for_temp_penalty = []
        weight_for_mask_increase = []
        for i in range(0,len(self.R_weight)):
            w_sum = self.R_weight[i].sum(dim=1,keepdim=True)
            temp2 = self.R_weight[i].clamp(min=0.001)
            temp = w_sum / w_sum.max()
            temp2 = temp2.pow(-1)
            temp2 = temp2.sum(dim=1,keepdim=True)
            temp2 = temp2 / temp2.max()
            weight_for_temp_penalty.append(temp)
            weight_for_mask_increase.append(temp2)
        return weight_for_temp_penalty[0] , weight_for_mask_increase[0]              


    def _obtain_current_result(self, step):
        """
        puts in self.current result the current result.
        also updates the best result
        :return:
        """
        if step == self.num_iter - 1 or step % 8 == 0:
            reflectance_out_np = np.clip(torch_to_np(self.reflectance_out), 0, 1)
            transmission_out_np = np.clip(torch_to_np(self.shading_out), 0, 1)
            
            psnr = compare_psnr(self.images[0],  reflectance_out_np * transmission_out_np)
            self.psnrs.append(psnr)
            self.current_result = SeparationResult(reflection=reflectance_out_np, transmission=transmission_out_np,
                                                   psnr=psnr)
        
            if self.best_result is None or self.best_result.psnr < self.current_result.psnr:
                self.best_result = self.current_result
        

    def finalize(self):
        save_graph('{}_PSNR'.format(self.image_name), self.psnrs, self.outputroot)
        save_image('{}_R_Final'.format(self.image_name), self.best_result.reflection, self.width,self.height,self.outputroot)
        save_image('{}_S_Final'.format(self.image_name), self.best_result.transmission, self.width,self.height,self.outputroot)
        save_image('{}_Original'.format(self.image_name), self.images[0], self.width,self.height,self.outputroot)

SeparationResult = namedtuple("SeparationResult", ['reflection', 'transmission' ,'psnr'])

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataroot', type=str, default='/home/jianghao/Project/Code/UIDNet-master/images_mit/', help='dataroot path')
    parser.add_argument('--outputroot', type=str, default='/home/jianghao/Project/Code/UIDNet-master/output_mit/', help='outputroot path')
    parser.add_argument('--param_mode', type=str, default='MIT', choices=['MIT', 'IIW', 'SAW'], help='choose a dataset to get better performed parameters')
    parser.add_argument('--gpu_ids', type=int, default=0, help='gpu index')
    opt = parser.parse_args()

    os.environ['CUDA_VISIBLE_DEVICES']='{}'.format(opt.gpu_ids) # set cuda device

    dataroot_I = opt.dataroot
    lines = os.listdir(dataroot_I)
    for i in range(len(lines)):
        if_srgb = True                                   
        eps = 1e-5
        img_name = lines[i]
        
        print('start processing image ', img_name)
        t1,R_weight, width,height = prepare_image_and_weight_v2(os.path.join(dataroot_I, img_name), if_srgb, eps, win_r=1, sigma_chro=0.025, sigma_I=0.1)
        s = Separation(opt, img_name[:-4], t1)       
        s.outputroot = opt.outputroot  
        s.R_weight = R_weight                 
        s.eps = eps

        s.R_skip = [0,0,64,128,0]                     
        s.S_skip = [0,0,64,128,0]
        s.R_pool_H = [2]*5                              
        s.R_pool_L = [2]*5                              

        s.smooth_start = 500
        s.start_pre = 0.01
        s.w_mask_l0 = 1; s.w_mask_R_l0 = 1
        s.w_mask_weight = 3; s.w_mask_R_weight = 3
        s.w_grad_l0 = 0

        s.metric = 'SSIM'

        if opt.param_mode == 'MIT':
            s.w_Rgrad_l0_comple = 0.02
            s.net_iters = 1
            s.mask_iters = 1
            s.R_mask_start = 100
            
            s.w_Rgrad_l0_ssim = 0.1
            s.w_Sgrad_l0 = 1
            s.ssim_win_size = 7
        else:
            s.w_Rgrad_l0_comple = 0
            s.net_iters = 10
            s.mask_iters = 20
            s.R_mask_start = 5000
        
            s.w_Rgrad_l0_ssim = 0.15
            s.w_Sgrad_l0 = 1
            s.ssim_win_size = 5

        s.loc_mean = 0
        s.beta = 0.1                    
        s.gamma = -0.1
        s.zeta = 1.1
        s.alpha = 1.2 
        s.ifdetach = True               

        s.S_pool_H = s.R_pool_H 
        s.S_pool_L = s.R_pool_L  

        s.w_grad = 0                           
        s.w_grad_smooth = 0
        s.threshold = 0.025
        s.learning_rate = 0.001                 
        s.num_iter = 10000
        if not os.path.exists(s.outputroot):
            os.makedirs(s.outputroot)
        if not os.path.exists(os.path.join(s.outputroot, img_name[-4:])):
            os.makedirs(os.path.join(s.outputroot, img_name[-4:]))
        s.outputroot = os.path.join(s.outputroot, img_name[-4:])
        s.width = width
        s.height = height
        s._init_all()
        s.optimize()
        s.finalize()       
        print('\nWe are done\n')
        
    
